
#include <propeller2.h>
#include <lut_exec.h>
#include <alloca.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>


//
// Prop2 boot pins, set in the Prop2 ROM
//
#define  MISO_EVAL  58
#define  MOSI_EVAL  59
#define  CS_EVAL  60
#define  CLK_EVAL  61

//
// Roger Loh's 4-bit add-on board
//
#define  BASE_PIN_RL  16
#define  DAT0_RL   BASE_PIN_RL+0    // MISO, DO
#define  DAT1_RL   BASE_PIN_RL+1
#define  DAT2_RL   BASE_PIN_RL+2
#define  DAT3_RL   BASE_PIN_RL+3    // CS
#define  CMD_RL    BASE_PIN_RL+4    // MOSI, DI
#define  CLK_RL    BASE_PIN_RL+5    // SCLK
#define  LED_RL    BASE_PIN_RL+6    // output to red LED
#define  PWR_RL    BASE_PIN_RL+7    // Card-Detect input, and power switch, and green LED

#define  CS_RL     DAT3_RL
#define  MOSI_RL   CMD_RL
#define  MISO_RL   DAT0_RL

//
// My hand wired full sized 4-bit SD slot
//
#define  BASE_PIN_EH  40
#define  CMD_EH    BASE_PIN_EH+2    // MOSI, DI
#define  CLK_EH    BASE_PIN_EH+3    // SCLK
#define  DAT0_EH   BASE_PIN_EH+4    // MISO, DO
#define  DAT1_EH   BASE_PIN_EH+5
#define  DAT2_EH   BASE_PIN_EH+6
#define  DAT3_EH   BASE_PIN_EH+7    // CS

#define  CS_EH     DAT3_EH
#define  MOSI_EH   CMD_EH
#define  MISO_EH   DAT0_EH




static uint32_t  randfill( uint32_t *addr, size_t size, uint32_t state )
{
    LUT_BEGIN("0", "randfill", "$FACE0000");    // load into lutRAM
    PASM(
"		shr	_PASM(size), #2\n"
"		wrfast	#0, _PASM(addr)\n"
"		rep	@.rend, _PASM(size)\n"
"		xoro32	_PASM(state)\n"
"		mov	pa, 0-0\n"
"		wflong	pa\n"
" .rend\n"
    );
    LUT_END;

    return state;    // for potential continuation of psuedo random sequence
}



static size_t  randcmp( const uint32_t *addr, size_t size, uint32_t state )
{
    LUT_BEGIN("0", "randcmp", "$FACE0001");    // load into lutRAM
    PASM(
"		shr	_PASM(size), #2\n"
"		mov	r0, #0\n"
"		rdfast	#0, _PASM(addr)\n"
"		rep	@.rend, _PASM(size)\n"
"		rflong	pa\n"
"		xoro32	_PASM(state)\n"
"		cmp	pa, 0-0   wz\n"
"	if_z	add	r0, #4\n"
" .rend\n"
    );
    return  LUT_END;
}



static void  emithex( uint8_t *addr, size_t size )
{
    do {
        printf(" %02x", *addr++);
    } while( --size );
    puts("");
}



static void  tester( const char *filename, const size_t kbytes, const size_t repeats )
{
    FILE  *fh;
    size_t  count;
    const size_t  bytes = kbytes * 1024;
    uint32_t  ticks, ticke, msecs, onesecond, seed, *buff;

    buff = alloca(bytes);    // auto-frees upon return from tester()
    if( !buff ) {    // auto-frees upon return from tester()
        printf(" malloc() failed!\r\n");
        exit(1);
    }

//-----------------------
// Write file
//-----------------------
    seed = _rnd();
    randfill(buff, bytes, seed);
    emithex((uint8_t *)buff, bytes);

    errno = 0;
    fh = fopen(filename, "wb");
    if( !fh ) {
        printf(" fopen() for writing failed!   errno = %d: %s\r\n", errno, strerror(errno));
        exit(3);
    }
    printf(" Buffer = %d kB, ", kbytes);

    onesecond = _clockfreq();
    msecs = 0;
    ticks = _cnt();
    count = repeats;
    do {
        if( bytes != fwrite(buff, bytes, 1, fh) )
            break;
        ticke = _cnt();
        while(ticke - ticks >= onesecond)  {    // one second chunk
            ticks += onesecond;    // one second in sysclock ticks
            msecs += 1000;    // one second in milliseconds
        }
    } while(--count);

    if( count ) {
        printf(" fwrite() failed!  writes = %d   errno = %d: %s\r\n", repeats - count, errno, strerror(errno));
        exit(4);
    }
    fclose(fh);

    ticke = _cnt();
    count = kbytes * repeats;
    msecs += _muldiv64(ticke - ticks, 1000, onesecond);  // include the remaining, ticks to milliseconds
    printf("Written %d kBytes at %d kB/s\r\n", count,
        (_muldiv64(count, 1000, msecs) + 512) / 1024);    // Bytes/s => kBytes/s

//-----------------------
// Read back file
//-----------------------
    errno = 0;
    fh = fopen(filename, "rb");
    if( !fh ) {
        printf(" fopen() for readback failed!   errno = %d: %s\r\n", errno, strerror(errno));
        exit(5);
    }
    count = repeats;
    do {
        if( bytes != fread(buff, 1, bytes, fh) )
            break;
        if( bytes != randcmp(buff, bytes, seed) )
            break;
    } while(--count);

    if( count )
        printf(" Mis-match! ");
    else
        printf(" Verified, ");

    rewind(fh);    // read the file again, this time purely for speed check
    errno = 0;
    onesecond = _clockfreq();
    msecs = 0;
    ticks = _cnt();
    count = repeats;
    do {
        if( bytes != fread(buff, 1, bytes, fh) )
            exit(6);
        ticke = _cnt();
        while(ticke - ticks >= onesecond)  {    // one second chunk
            ticks += onesecond;    // one second in sysclock ticks
            msecs += 1000;    // one second in milliseconds
        }
    } while(--count);
    fclose(fh);
    ticke = _cnt();
    count = kbytes * repeats;
    msecs += _muldiv64(ticke - ticks, 1000, onesecond);  // include the remaining, ticks to milliseconds
    printf("Read %d kBytes at %d kB/s\r\n", count,
        (_muldiv64(count, 1000, msecs) + 512) / 1024);    // Bytes/s => kBytes/s
}




static void  shutdown( void )
{
    puts("\r\nexit\r\n");  // blank line for re-run separation
    _waitms(500);
    _clkset(1, 20000);  // cool running
}



//===================================================================================

void  main( void )
{
    int  i;

    atexit(shutdown);
    printf("   clkfreq = %d   clkmode = 0x%x\r\n", _clockfreq(), _clockmode());

    puts("");
    for( i = 0; i < 3; i++ )
        tester( "speed1.bin", 2, 256 );    // 256 x 2 KB = 0.5 MB

    puts("");
    for( i = 0; i < 3; i++ )
        tester( "speed3.bin", 8, 256 );    // 256 x 8 KB = 2 MB

    puts("");
    for( i = 0; i < 3; i++ )
        tester( "speed5.bin", 32, 128 );    // 256 x 32 KB = 8 MB

    exit(0);
}
